from typing import List, Tuple

import asyncio
import trio
import serial
import serial_asyncio
from hedgehog.platform import Controller
from hedgehog.protocol.errors import FailedCommandError
from . import HardwareAdapter, POWER

# commands:
IO_STATE = 0x10
ANALOG_REQ = 0x20
DIGITAL_REQ = 0x30
MOTOR = 0x40
SERVO = 0x50
SERIAL = 0x60
# replies:
OK = 0x80
UNKNOWN_OPCODE = 0x81
INVALID_OPCODE = 0x82
INVALID_PORT = 0x83
INVALID_IO = 0x84
INVALID_MODE = 0x85
INVALID_FLAGS = 0x86
INVALID_VALUE = 0x87
ANALOG_REP = 0xA1
DIGITAL_REP = 0xB1
SERIAL_UPDATE = 0xE1
# number analog and digital ports together, servo and motor ports separately
# special analog ports:
BATTERY_VOLTAGE = 0x80
# special digital ports (output only):
LED1 = 0x90
LED2 = 0x91
BUZZER = 0x92
# serial ports:
SPI1 = 0x00
SPI2 = 0x01

_replies = {
    OK,
    ANALOG_REP,
    DIGITAL_REP,
}
_errors = {
    UNKNOWN_OPCODE,
    INVALID_OPCODE,
    INVALID_PORT,
    INVALID_IO,
    INVALID_MODE,
    INVALID_FLAGS,
    INVALID_VALUE,
}
_cmd_lengths = {
    OK: 1,
    UNKNOWN_OPCODE: 1,
    INVALID_OPCODE: 1,
    INVALID_PORT: 1,
    INVALID_IO: 1,
    INVALID_MODE: 1,
    INVALID_FLAGS: 1,
    INVALID_VALUE: 1,
    ANALOG_REP: 4,
    DIGITAL_REP: 3,
}


class TruncatedcommandError(FailedCommandError):
    pass


async def open_serial_connection(ser: serial.Serial, *,
                                 loop: asyncio.AbstractEventLoop=None, limit: int=asyncio.streams._DEFAULT_LIMIT
                                 ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
    if loop is None:
        loop = asyncio.get_event_loop()

    reader = asyncio.StreamReader(limit=limit, loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
    transport = serial_asyncio.SerialTransport(loop, protocol, ser)
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)
    return reader, writer


class SerialHardwareAdapter(HardwareAdapter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.controller = Controller()
        self.reader = None  # type: asyncio.StreamReader
        self.writer = None  # type: asyncio.StreamWriter
        self.controller.reset(True)

    async def __aenter__(self):
        await super().__aenter__()
        self.reader, self.writer = await open_serial_connection(self.controller.serial)
        # TODO register cleanup

    async def repeatable_command(self, cmd: List[int], reply_code: int=OK, tries: int=3) -> List[int]:
        for i in range(tries - 1):
            try:
                return await self.command(cmd, reply_code=reply_code)
            except TruncatedcommandError:
                pass
        return await self.command(cmd, reply_code=reply_code)

    async def command(self, cmd: List[int], reply_code: int=OK) -> List[int]:
        async def read_command() -> List[int]:
            cmd = await self.reader.read(1)
            if cmd[0] in _cmd_lengths:
                length = _cmd_lengths[cmd[0]]
                if length > 1:
                    cmd += await self.reader.read(length - 1)
            else:
                cmd += await self.reader.read(1)
                length = cmd[1] + 2
                if length > 2:
                    cmd += await self.reader.read(length - 2)
            if len(cmd) != length:
                raise TruncatedcommandError("HWC sent a truncated response")
            return list(cmd)

        self.writer.write(bytes(cmd))
        reply = await read_command()
        while reply[0] not in _replies | _errors:
            # TODO do something with the update
            reply = await read_command()

        if reply[0] in _replies:
            assert reply[0] == reply_code
            return reply
        else:
            if reply[0] == UNKNOWN_OPCODE:
                self.controller.serial.flushOutput()
                await trio.sleep(0.02)
                self.controller.serial.flushInput()
                raise FailedCommandError("opcode unknown to the HWC; connection was reset")
            elif reply[0] == INVALID_OPCODE:
                raise FailedCommandError("opcode not supported")
            elif reply[0] == INVALID_PORT:
                raise FailedCommandError("port not supported or out of range")
            elif reply[0] == INVALID_IO:
                raise FailedCommandError("sensor request invalid for output port")
            elif reply[0] == INVALID_MODE:
                raise FailedCommandError("unsupported motor mode")
            elif reply[0] == INVALID_FLAGS:
                raise FailedCommandError("unsupported combination of IO flags")
            elif reply[0] == INVALID_VALUE and cmd[0] == MOTOR:
                raise FailedCommandError("unsupported motor power/velocity")
            elif reply[0] == INVALID_VALUE and cmd[0] == SERVO:
                raise FailedCommandError("unsupported servo position")
            raise FailedCommandError("unknown hardware controller response")

    async def set_io_state(self, port, flags):
        await self.repeatable_command([IO_STATE, port, flags])

    async def get_analog(self, port):
        _, port_, value_hi, value_lo = await self.repeatable_command([ANALOG_REQ, port], ANALOG_REP)
        assert port_ == port
        return int.from_bytes([value_hi, value_lo], 'big')

    async def get_digital(self, port):
        _, port_, value = await self.repeatable_command([DIGITAL_REQ, port], DIGITAL_REP)
        assert port_ == port
        assert value & ~0x01 == 0x00
        return value != 0

    async def set_motor(self, port, state, amount=0, reached_state=POWER, relative=None, absolute=None):
        if not -0x8000 < amount < 0x8000:
            raise FailedCommandError("unsupported motor power/velocity")
        value = amount if amount > 0 else (0x8000 | -amount)
        value_hi, value_lo = value.to_bytes(2, 'big')
        await self.repeatable_command([MOTOR, port, state, value_hi, value_lo])

    async def set_servo(self, port, active, position):
        if not 0 <= position < 0x8000:
            raise FailedCommandError("unsupported servo position")
        value = position | (0x8000 if active else 0x0000)
        value_hi, value_lo = value.to_bytes(2, 'big')
        await self.repeatable_command([SERVO, port, value_hi, value_lo])



from typing import Dict

import random

from . import HardwareAdapter, POWER
from hedgehog.protocol.messages import io


class SimulatedHardwareAdapter(HardwareAdapter):
    def __init__(self, *args, simulate_sensors=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.simulate_sensors = simulate_sensors

        self.io_states = {}  # type: Dict[int, int]

    async def set_io_state(self, port, flags):
        self.io_states[port] = flags

    async def get_analog(self, port):
        if not self.simulate_sensors:
            return 0

        mu, sigma = {
            io.INPUT_FLOATING: (800, 60),
            io.INPUT_PULLUP: (4030, 30),
            io.INPUT_PULLDOWN: (80, 30),
            io.OUTPUT_ON: (4050, 20),
            io.OUTPUT_OFF: (50, 20),
        }[self.io_states.get(port, io.INPUT_FLOATING)]

        num = int(random.gauss(mu, sigma))
        if num < 0:
            num = 0
        if num >= 4096:
            num = 4095
        return num

    async def get_digital(self, port):
        if not self.simulate_sensors:
            return False

        value = {
            io.INPUT_FLOATING: False,
            io.INPUT_PULLUP: True,
            io.INPUT_PULLDOWN: False,
            io.OUTPUT_ON: True,
            io.OUTPUT_OFF: False,
        }[self.io_states.get(port, io.INPUT_FLOATING)]
        return value

    async def set_motor(self, port, state, amount=0, reached_state=POWER, relative=None, absolute=None):
        # TODO set motor action
        pass

    async def get_motor(self, port):
        return 0, 0

    async def set_motor_position(self, port, position):
        # TODO set motor position
        pass

    async def set_servo(self, port, active, position):
        # TODO set servo position
        pass




from typing import Tuple

from contextlib import AsyncExitStack
from dataclasses import dataclass
import trio

from hedgehog.protocol.errors import UnsupportedCommandError
from hedgehog.protocol import messages
from hedgehog.protocol.messages import io, analog, digital, servo
from hedgehog.protocol.messages.motor import POWER


class HardwareUpdate:
    pass


@dataclass
class MotorStateUpdate(HardwareUpdate):
    port: int
    state: int


class HardwareAdapter(object):
    def __init__(self) -> None:
        self._send_channel, self.hardware_updates = trio.open_memory_channel(10)
        self._stack: AsyncExitStack = None

    async def __aenter__(self):
        self._stack = AsyncExitStack()
        await self._stack.enter_async_context(self._send_channel)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self._stack.__aexit__(exc_type, exc_val, exc_tb)

    def _enqueue_update(self, update: HardwareUpdate):
        try:
            # try enqueueing the update
            self._send_channel.send_nowait(update)
        except trio.WouldBlock:  # pragma: nocover
            try:
                # the queue is full, so this must work
                self.hardware_updates.receive_nowait()
            except trio.WouldBlock:
                assert False
            try:
                # now the queue can't be full, try again
                self._send_channel.send_nowait(update)
            except trio.WouldBlock:
                assert False

    async def set_io_state(self, port: int, flags: int) -> None:
        raise UnsupportedCommandError(messages.io.Action.msg_name())

    async def get_analog(self, port: int) -> int:
        raise UnsupportedCommandError(messages.analog.Request.msg_name())

    async def get_digital(self, port: int) -> bool:
        raise UnsupportedCommandError(messages.digital.Request.msg_name())

    async def set_motor(self, port: int, state: int, amount: int=0,
                  reached_state: int=POWER, relative: int=None, absolute: int=None) -> None:
        raise UnsupportedCommandError(messages.motor.Action.msg_name())

    async def get_motor(self, port: int) -> Tuple[int, int]:
        raise UnsupportedCommandError(messages.motor.StateRequest.msg_name())

    def motor_state_update(self, port: int, state: int) -> None:
        self._enqueue_update(MotorStateUpdate(port, state))

    async def set_motor_position(self, port: int, position: int) -> None:
        raise UnsupportedCommandError(messages.motor.SetPositionAction.msg_name())

    async def set_servo(self, port: int, active: bool, position: int) -> None:
        raise UnsupportedCommandError(messages.servo.Action.msg_name())



import functools
from typing import Awaitable, Callable, Dict, Type

from hedgehog.protocol import Header, Message
from hedgehog.utils import SimpleDecorator, Registry
from ..hedgehog_server import HedgehogServer


HandlerFunction = Callable[['CommandHandler', HedgehogServer, Header, Message], Awaitable[Message]]
HandlerCallback = Callable[[HedgehogServer, Header, Message], Awaitable[Message]]
HandlerCallbackDict = Dict[Type[Message], HandlerCallback]
HandlerDecorator = Callable[[Type[Message]], SimpleDecorator[HandlerFunction]]


CommandRegistry = Registry[Type[Message], HandlerFunction]


class CommandHandler(object):
    _commands = None  # type: CommandRegistry

    @property
    def handlers(self) -> HandlerCallbackDict:
        return {
            key: functools.partial(handler, self)
            for key, handler in self._commands.items()
        }


def merge(*handlers: CommandHandler) -> HandlerCallbackDict:
    result = {}  # type: HandlerCallbackDict
    for handler in handlers:
        dups = result.keys() & handler.handlers.keys()
        if len(dups) > 0:
            raise ValueError("Duplicate command handler for {}".format([dup.msg_name() for dup in dups]))
        result.update(handler.handlers)
    return result



from typing import AsyncIterator, Dict

import asyncio.subprocess
import trio
import trio_asyncio

from hedgehog.protocol.errors import FailedCommandError
from hedgehog.protocol.messages import ack, process, motor

from . import CommandHandler, CommandRegistry
from ..hedgehog_server import Job
from ..hardware import HardwareAdapter


class ProcessHandler(CommandHandler):
    _commands = CommandRegistry()

    def __init__(self, adapter: HardwareAdapter) -> None:
        super().__init__()
        self._processes = {}  # type: Dict[int, asyncio.subprocess.Process]
        self.adapter = adapter

    async def _handle_process(self, server, ident, msg, *, task_status=trio.TASK_STATUS_IGNORED) -> AsyncIterator[Job]:
        proc = await trio_asyncio.aio_as_trio(asyncio.create_subprocess_exec)(
            *msg.args,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=msg.working_dir)

        pid = proc.pid
        self._processes[pid] = proc

        task_status.started(pid)

        async with trio.open_nursery() as nursery:
            async def handle_stream(fileno, file):
                while True:
                    chunk = await trio_asyncio.aio_as_trio(file.read)(4096)
                    async with server.job():
                        await server.send_async(ident, process.StreamUpdate(pid, fileno, chunk))
                    if chunk == b'':
                        break

            nursery.start_soon(handle_stream, process.STDOUT, proc.stdout)
            nursery.start_soon(handle_stream, process.STDERR, proc.stderr)

        exit_code = await trio_asyncio.aio_as_trio(proc.wait)()

        try:
            # turn off all actuators
            # TODO hard coded number of ports
            for port in range(4):
                async with server.job():
                    await self.adapter.set_motor(port, motor.POWER, 0)
            for port in range(4):
                async with server.job():
                    await self.adapter.set_servo(port, False, 0)
        except Exception:  # pragma: nocover
            async with server.job():
                await server.send_async(ident, process.ExitUpdate(pid, exit_code))
            raise
        else:
            async with server.job():
                await server.send_async(ident, process.ExitUpdate(pid, exit_code))
        finally:
            del self._processes[pid]

    @_commands.register(process.ExecuteAction)
    async def process_execute_action(self, server, ident, msg):
        pid = await server.add_task(self._handle_process, server, ident, msg)
        return process.ExecuteReply(pid)

    @_commands.register(process.StreamAction)
    async def process_stream_action(self, server, ident, msg):
        # check whether the process has already finished
        if msg.pid in self._processes:
            assert msg.fileno == process.STDIN

            proc = self._processes[msg.pid]
            if msg.chunk != b'':
                proc.stdin.write(msg.chunk)
            else:
                proc.stdin.write_eof()
            return ack.Acknowledgement()
        else:
            raise FailedCommandError("no process with pid {}".format(msg.pid))

    @_commands.register(process.SignalAction)
    async def process_signal_action(self, server, ident, msg):
        # check whether the process has already finished
        if msg.pid in self._processes:
            proc = self._processes[msg.pid]
            proc.send_signal(msg.signal)
            return ack.Acknowledgement()
        else:
            raise FailedCommandError("no process with pid {}".format(msg.pid))



import argparse
import configparser
import logging
import logging.config
import os.path
import subprocess
import trio
import trio_asyncio
from contextlib import suppress

from hedgehog.utils.zmq import trio as zmq_trio

from . import handlers
from .hedgehog_server import HedgehogServer
from .handlers.hardware import HardwareHandler
from .handlers.process import ProcessHandler

logger = logging.getLogger(__name__)


def parse_args(simulator=False):
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--port', type=int, default=0,
                        help="The port to use, 0 means random port; default: %(default)s")
    parser.add_argument('--scan', '--scan-config', dest='scan_config', action='store_true',
                        help="If given, a config file will be processed at startup if it exists")
    parser.add_argument('--scan-file', '--scan-config-file', dest='scan_config_file', default='/media/usb/hedgehog.conf',
                        help="Location of the config file to scan; default: %(default)s")
    parser.add_argument('--config-file', dest='config_file', default='hedgehog.conf',
                        help="The hedgehog config file; default: %(default)s")
    parser.add_argument('--logging-conf', dest='logging_conf',
                        help="If given, logging is configured from this file")
    if simulator:
        parser.add_argument('--sensors', '--simulate-sensors', dest='simulate_sensors', action='store_true',
                            help="If given, noisy sensor readings will be simulated")
    return parser.parse_args()


def apply_scan_config(config, scan_config):
    def set(config, section, option, value):
        if value is not None:
            if section not in config:
                config.add_section(section)
            config[section][option] = value
        elif section in config and option in config[section]:
            del config[section][option]
            if not config[section]:
                del config[section]

    def get(config, section, option):
        return config.get(section, option, fallback=None)

    def copy(in_cfg, out_cfg, section, option):
        value = get(in_cfg, section, option)
        set(out_cfg, section, option, value)
        return value

    copy(scan_config, config, 'default', 'name')
    copy(scan_config, config, 'default', 'port')
    copy(scan_config, config, 'default', 'services')
    wifi_commands = get(scan_config, 'wifi', 'commands')

    if wifi_commands:
        # TODO make this less dangerous when called on a non-hedgehog machine
        subprocess.Popen(['wpa_cli'], stdin=subprocess.PIPE).communicate(wifi_commands.encode())


def launch(hardware):
    from hedgehog.server.hardware.simulated import SimulatedHardwareAdapter
    simulator = hardware == SimulatedHardwareAdapter

    args = parse_args(simulator)

    if args.logging_conf:
        logging.config.fileConfig(args.logging_conf)

    if simulator and args.simulate_sensors:
        _hardware = hardware

        def hardware(*args, **kwargs):
            return _hardware(*args, simulate_sensors=True, **kwargs)

    config = configparser.ConfigParser()
    config.read(args.config_file)

    if args.scan_config and os.path.isfile(args.scan_config_file):
        scan_config = configparser.ConfigParser()
        scan_config.read(args.scan_config_file)

        apply_scan_config(config, scan_config)

        with open(args.config_file, mode='w') as f:
            config.write(f)

    port = args.port or config.getint('default', 'port', fallback=0)

    with suppress(KeyboardInterrupt):
        start(hardware, port)


def start(hardware, port=0):
    ctx = zmq_trio.Context.instance()

    async def run():
        adapter = hardware()
        handler = handlers.merge(HardwareHandler(adapter), ProcessHandler(adapter))

        async with trio_asyncio.open_loop(), adapter:
            await HedgehogServer(ctx, 'tcp://*:{}'.format(port), handler).run()

    trio.run(run)



from typing import Any, Dict, Generic, List, Set, Tuple, TypeVar

import trio
import bisect

from . import HardwareAdapter, HardwareUpdate, POWER


T = TypeVar('T')


class MockedState(Generic[T]):
    def __init__(self) -> None:
        self._times: List[float] = []
        self._values: List[T] = []

    def set(self, time: float, value: T) -> None:
        i = bisect.bisect_left(self._times, time)
        if i < len(self._times) and self._times[i] == time:
            self._values[i] = value
        else:
            self._times.insert(i, time)
            self._values.insert(i, value)

    def get(self, time: float=None, default: T=None)-> T:
        if time is None:
            time = trio.current_time()

        i = bisect.bisect_right(self._times, time)
        if i == 0:
            return default
        return self._values[i - 1]


class MockedUpdates:
    def __init__(self) -> None:
        self._updates: Dict[float, Set[HardwareUpdate]] = {}

    def add(self, time: float, update: HardwareUpdate) -> None:
        try:
            updates = self._updates[time]
        except KeyError:
            updates = self._updates[time] = set()
        updates.add(update)

    async def __aiter__(self):
        while self._updates:
            next_time = min(self._updates)
            await trio.sleep_until(next_time)
            updates = self._updates.pop(next_time)
            for update in updates:
                yield update


class MockedHardwareAdapter(HardwareAdapter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.io_states: Dict[int, int] = {}
        self._updates: MockedUpdates = MockedUpdates()
        self._analogs: List[MockedState[int]] = [MockedState() for port in range(16)]
        self._digitals: List[MockedState[bool]] = [MockedState() for port in range(16)]
        self._motors: List[MockedState[Tuple[float, float]]] = [MockedState() for port in range(4)]

    async def __aenter__(self):
        await super().__aenter__()
        nursery = await self._stack.enter_async_context(trio.open_nursery())

        @nursery.start_soon
        async def emit_updates():
            async with self._send_channel:
                async for update in self._updates:
                    self._enqueue_update(update)

    async def set_io_state(self, port, flags):
        self.io_states[port] = flags

    def set_analog(self, port: int, time: float, value: int) -> None:
        self._analogs[port].set(time, value)

    async def get_analog(self, port):
        return self._analogs[port].get(default=0)

    def set_digital(self, port: int, time: float, value: bool) -> None:
        self._digitals[port].set(time, value)

    async def get_digital(self, port):
        return self._digitals[port].get(default=False)

    async def set_motor(self, port, state, amount=0, reached_state=POWER, relative=None, absolute=None):
        # TODO set motor action
        pass

    def set_motor_state(self, port: int, time: float, velocity: int, position: int) -> None:
        self._motors[port].set(time, (velocity, position))

    async def get_motor(self, port):
        return self._motors[port].get(default=(0, 0))

    async def set_motor_position(self, port, position):
        # TODO set motor position
        pass

    async def set_servo(self, port, active, position):
        # TODO set servo position
        pass




from typing import AsyncIterator, Awaitable, Callable, Dict, Type

import logging
import trio
import zmq
from contextlib import asynccontextmanager
from functools import partial
from hedgehog.utils.zmq import trio as zmq_trio
from hedgehog.protocol import ServerSide, Header, RawMessage, Message, RawPayload
from hedgehog.protocol.zmq.trio import DealerRouterSocket
from hedgehog.protocol.errors import HedgehogCommandError, UnsupportedCommandError, FailedCommandError


# TODO importing this from .handlers does not work...
HandlerCallback = Callable[['HedgehogServer', Header, Message], Awaitable[Message]]
Job = Callable[[], Awaitable[None]]

logger = logging.getLogger(__name__)


class HedgehogServer:
    def __init__(self, ctx: zmq_trio.Context, endpoint: str, handlers: Dict[Type[Message], HandlerCallback]) -> None:
        self.ctx = ctx
        self._nursery = None
        self._lock = trio.StrictFIFOLock()
        self.endpoint = endpoint
        self.handlers = handlers
        self.socket: DealerRouterSocket = None

    def stop(self):
        self._nursery.cancel_scope.cancel()

    async def _requests_task(self, *, task_status=trio.TASK_STATUS_IGNORED) -> None:
        async def handle_msg(ident: Header, msg_raw: RawMessage) -> RawMessage:
            try:
                msg = ServerSide.parse(msg_raw)
                logger.debug("Receive command: %s", msg)
                try:
                    handler = self.handlers[msg.__class__]
                except KeyError:
                    raise UnsupportedCommandError(msg.__class__.msg_name())
                try:
                    result = await handler(self, ident, msg)
                except HedgehogCommandError:
                    raise
                except Exception as err:
                    logger.exception("Uncaught exception in command handler")
                    raise FailedCommandError("Uncaught exception: {}".format(repr(err))) from err
            except HedgehogCommandError as err:
                result = err.to_message()
            logger.debug("Send reply:      %s", result)
            return ServerSide.serialize(result)

        task_status.started()
        while True:
            ident, msgs_raw = await self.socket.recv_msgs_raw()
            async with self.job():
                await self.socket.send_msgs_raw(ident, [await handle_msg(ident, msg) for msg in msgs_raw])

    async def add_task(self, async_fn, *args, name=None):
        async def async_fn_wrapper(*, task_status=trio.TASK_STATUS_IGNORED) -> None:
            logger.debug("Added new task: %s", name if name else async_fn)
            try:
                await async_fn(*args, task_status=task_status)
            except Exception:
                logger.exception("Task raised an exception: %s", name if name else async_fn)
            else:
                logger.debug("Task finished: %s", name if name else async_fn)

        return await self._nursery.start(async_fn_wrapper)

    @asynccontextmanager
    async def job(self) -> None:
        LONG_RUNNING_THRESHOLD = 0.1
        CANCEL_THRESHOLD = 10

        async with self._lock:
            job = None  # TODO identify the job
            nursery = None
            manually_cancelled = False
            begin = trio.current_time()
            try:
                async with trio.open_nursery() as nursery:
                    nursery.cancel_scope.deadline = begin + CANCEL_THRESHOLD

                    @nursery.start_soon
                    async def warn_long_running():
                        await trio.sleep(LONG_RUNNING_THRESHOLD)
                        logger.warning("Long running job on server loop: %s", job)

                    yield

                    # cancel the warning task
                    manually_cancelled = True
                    nursery.cancel_scope.cancel()
            finally:
                assert nursery is not None

                end = trio.current_time()
                if nursery.cancel_scope.cancelled_caught and not manually_cancelled:
                    logger.error("Long running job cancelled after %.1f ms: %s", (end - begin) * 1000, job)
                    raise trio.TooSlowError
                elif end - begin > LONG_RUNNING_THRESHOLD:
                    logger.warning("Long running job finished after %.1f ms: %s", (end - begin) * 1000, job)

    async def send_async(self, ident: Header, *msgs: Message) -> None:
        for msg in msgs:
            logger.debug("Send update:     %s", msg)
        await self.socket.send_msgs(ident, msgs)

    async def run(self, *, task_status=trio.TASK_STATUS_IGNORED):
        with DealerRouterSocket(self.ctx, zmq.ROUTER, side=ServerSide) as self.socket:
            async with trio.open_nursery() as self._nursery:
                self.socket.bind(self.endpoint)
                task_status.started()

                await self._nursery.start(self._requests_task)
        logger.info("Server stopped")



from typing import cast, Dict, Tuple, Type

import itertools
from hedgehog.protocol import Header, Message
from hedgehog.protocol.errors import FailedCommandError, UnsupportedCommandError
from hedgehog.protocol.messages import ack, io, analog, digital, motor, servo
from hedgehog.protocol.proto.subscription_pb2 import Subscription

from . import CommandHandler, CommandRegistry
from .. import subscription
from ..hardware import HardwareAdapter
from ..hedgehog_server import HedgehogServer


class _HWHandler(object):
    def __init__(self, adapter: HardwareAdapter) -> None:
        self.adapter = adapter
        self.subscribables = {}  # type: Dict[Type[Message], subscription.Subscribable]

    async def subscribe(self, server: HedgehogServer, ident: Header, msg: Type[Message], subscription: Subscription) -> None:
        try:
            subscribable = self.subscribables[msg]
        except KeyError as err:  # pragma: nocover
            raise UnsupportedCommandError(msg.msg_name()) from err
        else:
            await subscribable.subscribe(server, ident, subscription)


class _IOHandler(_HWHandler):
    def __command_subscribable(self) -> subscription.TriggeredSubscribable[int, io.CommandUpdate]:
        outer_self = self

        class Subs(subscription.TriggeredSubscribable[int, io.CommandUpdate]):
            def compose_update(self, server, ident, subscription, flags):
                return io.CommandUpdate(outer_self.port, flags, subscription)

        return Subs()

    def __analog_subscribable(self) -> subscription.PolledSubscribable[int, analog.Update]:
        outer_self = self

        class Subs(subscription.PolledSubscribable[int, analog.Update]):
            async def poll(self):
                return await outer_self.analog_value

            def compose_update(self, server, ident, subscription, value):
                return analog.Update(outer_self.port, value, subscription)

        return Subs()

    def __digital_subscribable(self) -> subscription.PolledSubscribable[bool, digital.Update]:
        outer_self = self

        class Subs(subscription.PolledSubscribable[bool, digital.Update]):
            async def poll(self):
                return await outer_self.digital_value

            def compose_update(self, server, ident, subscription, value):
                return digital.Update(outer_self.port, value, subscription)

        return Subs()

    def __init__(self, adapter: HardwareAdapter, port: int) -> None:
        super(_IOHandler, self).__init__(adapter)
        self.port = port
        self.command = None  # type: Tuple[int]
        self.subscribables[io.CommandSubscribe] = self.__command_subscribable()
        self.subscribables[analog.Subscribe] = self.__analog_subscribable()
        self.subscribables[digital.Subscribe] = self.__digital_subscribable()

    async def action_update(self) -> None:
        if self.command is None:
            return
        flags, = self.command
        await cast(subscription.TriggeredSubscribable[int, io.CommandUpdate],
                   self.subscribables[io.CommandSubscribe]).update(flags)

    async def action(self, flags: int) -> None:
        await self.adapter.set_io_state(self.port, flags)
        self.command = flags,
        await self.action_update()

    @property
    async def analog_value(self) -> int:
        return await self.adapter.get_analog(self.port)

    @property
    async def digital_value(self) -> bool:
        return await self.adapter.get_digital(self.port)


class _MotorHandler(_HWHandler):
    def __command_subscribable(self) -> subscription.TriggeredSubscribable[Tuple[motor.Config, int, int], motor.CommandUpdate]:
        outer_self = self

        class Subs(subscription.TriggeredSubscribable[Tuple[motor.Config, int, int], motor.CommandUpdate]):
            def compose_update(self, server, ident, subscription, command):
                config, state, amount = command
                return motor.CommandUpdate(outer_self.port, config, state, amount, subscription)

        return Subs()

    def __state_subscribable(self) -> subscription.PolledSubscribable[Tuple[int, int], motor.StateUpdate]:
        outer_self = self

        class Subs(subscription.PolledSubscribable[Tuple[int, int], motor.StateUpdate]):
            async def poll(self):
                return await outer_self.state

            def compose_update(self, server, ident, subscription, state):
                velocity, position = state
                return motor.StateUpdate(outer_self.port, velocity, position, subscription)

        return Subs()

    def __init__(self, adapter: HardwareAdapter, port: int) -> None:
        super(_MotorHandler, self).__init__(adapter)
        self.port = port
        self.config = motor.DcConfig()  # type: motor.Config
        self.command = None  # type: Tuple[int, int]

        self.subscribables[motor.CommandSubscribe] = self.__command_subscribable()
        try:
            # TODO
            # self.state
            pass
        except UnsupportedCommandError:  # pragma: nocover
            pass
        else:
            self.subscribables[motor.StateSubscribe] = self.__state_subscribable()

    async def action_update(self) -> None:
        if self.command is None:
            return
        state, amount = self.command
        await cast(subscription.TriggeredSubscribable[Tuple[int, int], motor.CommandUpdate],
                   self.subscribables[motor.CommandSubscribe]).update((self.config, state, amount))

    async def action(self, state: int, amount: int, reached_state: int, relative: int, absolute: int) -> None:
        await self.adapter.set_motor(self.port, state, amount, reached_state, relative, absolute)
        self.command = state, amount
        await self.action_update()

    async def config_action(self, config: motor.Config) -> None:
        # await self.adapter.set_motor(self.port, state, amount, reached_state, relative, absolute)
        self.config = config
        await self.action_update()

    async def set_position(self, position: int) -> None:
        await self.adapter.set_motor_position(self.port, position)

    @property
    async def state(self) -> Tuple[int, int]:
        return await self.adapter.get_motor(self.port)


class _ServoHandler(_HWHandler):
    def __command_subscribable(self) -> subscription.TriggeredSubscribable[Tuple[int, int], servo.CommandUpdate]:
        outer_self = self

        class Subs(subscription.TriggeredSubscribable[Tuple[int, int], servo.CommandUpdate]):
            def compose_update(self, server, ident, subscription, command):
                active, position = command
                return servo.CommandUpdate(outer_self.port, active, position, subscription)

        return Subs()

    def __init__(self, adapter: HardwareAdapter, port: int) -> None:
        super(_ServoHandler, self).__init__(adapter)
        self.port = port
        self.command = None  # type: Tuple[bool, int]

        self.subscribables[servo.CommandSubscribe] = self.__command_subscribable()

    async def action_update(self) -> None:
        if self.command is None:
            return
        active, position = self.command
        await cast(subscription.TriggeredSubscribable[Tuple[int, int], servo.CommandUpdate],
                   self.subscribables[servo.CommandSubscribe]).update((active, position))

    async def action(self, active: bool, position: int) -> None:
        await self.adapter.set_servo(self.port, active, position if active else 0)
        self.command = active, position
        await self.action_update()


class HardwareHandler(CommandHandler):
    _commands = CommandRegistry()

    def __init__(self, adapter: HardwareAdapter) -> None:
        super().__init__()
        self.adapter = adapter
        # TODO hard-coded number of ports
        self.ios = {port: _IOHandler(adapter, port) for port in itertools.chain(range(0, 16), (0x80, 0x90, 0x91))}
        self.motors = [_MotorHandler(adapter, port) for port in range(0, 4)]
        self.servos = [_ServoHandler(adapter, port) for port in range(0, 6)]
        # self.motor_cb = {}
        # self.adapter.motor_state_update_cb = self.motor_state_update

    @_commands.register(io.Action)
    async def io_state_action(self, server, ident, msg):
        await self.ios[msg.port].action(msg.flags)
        return ack.Acknowledgement()

    @_commands.register(io.CommandRequest)
    async def io_command_request(self, server, ident, msg):
        command = self.ios[msg.port].command
        try:
            flags, = command
        except TypeError:
            raise FailedCommandError("no command executed yet")
        else:
            return io.CommandReply(msg.port, flags)

    @_commands.register(io.CommandSubscribe)
    async def io_command_subscribe(self, server, ident, msg):
        await self.ios[msg.port].subscribe(server, ident, msg.__class__, msg.subscription)
        await self.ios[msg.port].action_update()
        return ack.Acknowledgement()

    @_commands.register(analog.Request)
    async def analog_request(self, server, ident, msg):
        value = await self.ios[msg.port].analog_value
        return analog.Reply(msg.port, value)

    @_commands.register(analog.Subscribe)
    async def analog_subscribe(self, server, ident, msg):
        await self.ios[msg.port].subscribe(server, ident, msg.__class__, msg.subscription)
        return ack.Acknowledgement()

    @_commands.register(digital.Request)
    async def digital_request(self, server, ident, msg):
        value = await self.ios[msg.port].digital_value
        return digital.Reply(msg.port, value)

    @_commands.register(digital.Subscribe)
    async def digital_subscribe(self, server, ident, msg):
        await self.ios[msg.port].subscribe(server, ident, msg.__class__, msg.subscription)
        return ack.Acknowledgement()

    @_commands.register(motor.Action)
    async def motor_action(self, server, ident, msg):
        # if msg.relative is not None or msg.absolute is not None:
        #     # this action will end with a state update
        #     def cb(port, state):
        #         server.send_async(ident, motor.StateUpdate(port, state))
        #     self.motor_cb[msg.port] = cb
        await self.motors[msg.port].action(msg.state, msg.amount, msg.reached_state, msg.relative, msg.absolute)
        return ack.Acknowledgement()

    @_commands.register(motor.ConfigAction)
    async def motor_config_action(self, server, ident, msg):
        await self.motors[msg.port].config_action(msg.config)
        return ack.Acknowledgement()

    @_commands.register(motor.CommandRequest)
    async def motor_command_request(self, server, ident, msg):
        config = self.motors[msg.port].config
        command = self.motors[msg.port].command
        try:
            state, amount = command
        except TypeError:
            raise FailedCommandError("no command executed yet")
        else:
            return motor.CommandReply(msg.port, config, state, amount)

    @_commands.register(motor.CommandSubscribe)
    async def motor_command_subscribe(self, server, ident, msg):
        await self.motors[msg.port].subscribe(server, ident, msg.__class__, msg.subscription)
        await self.motors[msg.port].action_update()
        return ack.Acknowledgement()

    @_commands.register(motor.StateRequest)
    async def motor_state_request(self, server, ident, msg):
        velocity, position = await self.motors[msg.port].state
        return motor.StateReply(msg.port, velocity, position)

    @_commands.register(motor.StateSubscribe)
    async def motor_state_subscribe(self, server, ident, msg):
        await self.motors[msg.port].subscribe(server, ident, msg.__class__, msg.subscription)
        return ack.Acknowledgement()

    # async def motor_state_update(self, port, state):
    #     if port in self.motor_cb:
    #         self.motor_cb[port](port, state)
    #         del self.motor_cb[port]

    @_commands.register(motor.SetPositionAction)
    async def motor_set_position_action(self, server, ident, msg):
        await self.motors[msg.port].set_position(msg.position)
        return ack.Acknowledgement()

    @_commands.register(servo.Action)
    async def servo_action(self, server, ident, msg):
        await self.servos[msg.port].action(msg.active, msg.position)
        return ack.Acknowledgement()

    @_commands.register(servo.CommandRequest)
    async def servo_command_request(self, server, ident, msg):
        command = self.servos[msg.port].command
        try:
            active, position = command
        except TypeError:
            raise FailedCommandError("no command executed yet")
        else:
            return servo.CommandReply(msg.port, active, position)

    @_commands.register(servo.CommandSubscribe)
    async def servo_command_subscribe(self, server, ident, msg):
        await self.servos[msg.port].subscribe(server, ident, msg.__class__, msg.subscription)
        await self.servos[msg.port].action_update()
        return ack.Acknowledgement()



from typing import AsyncIterator, Callable, Dict, Generic, Set, TypeVar

import math
import trio
from contextlib import AsyncExitStack, asynccontextmanager
from functools import partial

from hedgehog.protocol import Header
from hedgehog.protocol.proto.subscription_pb2 import Subscription
from hedgehog.protocol.errors import FailedCommandError

from .hedgehog_server import HedgehogServer


T = TypeVar('T')
Upd = TypeVar('Upd')


class BroadcastChannel(Generic[T], trio.abc.AsyncResource):
    """\
    Bundles a set of trio channels so that messages are sent to all of them.
    When a receiver is closed, it is cleanly removed from the broadcast channel on the next send.
    Be careful about the buffer size chosen when adding a receiver; see `send()` for details.
    """

    def __init__(self) -> None:
        self._send_channels: Set[trio.abc.SendChannel] = set()
        self._stack = AsyncExitStack()

    async def send(self, value: T) -> None:
        """\
        Sends the value to all receiver channels.
        Closed receivers are removed the next time a value.'is sent using this method.
        This method will send to all receivers immediately,
        but it will block until the message got out to all receivers.

        Suppose you have receivers A and B with buffer size zero, and you send to them:

            await channel.send(1)
            await channel.send(2)

        If only B is actually reading, then `send(2)` will not be called, because `send(1)` can't finish,
        meaning the `2` is not delivered to B either.
        To prevent this, close any receivers that are done, and/or poll receive in a timely manner.
        """
        broken = set()

        async def send(channel):
            try:
                await channel.send(value)
            except trio.BrokenResourceError:
                await channel.aclose()
                broken.add(channel)

        async with trio.open_nursery() as nursery:
            for channel in self._send_channels:
                nursery.start_soon(send, channel)

        self._send_channels -= broken
        broken.clear()

    def add_receiver(self, max_buffer_size) -> trio.abc.ReceiveChannel:
        """\
        Adds a receiver to this broadcast channel with the given buffer capacity.
        The send end of the receiver is closed when the broadcast channel is closed,
        and if the receive end is closed, it is discarded from the broadcast channel.
        """
        send, receive = trio.open_memory_channel(max_buffer_size)
        self._stack.push_async_exit(send)
        self._send_channels.add(send)
        return receive

    async def aclose(self):
        """\
        Closes the broadcast channel, causing all receive channels to stop iteration.
        """
        await self._stack.aclose()


@asynccontextmanager
async def _skipping_stream(stream: AsyncIterator[T]):
    """\
    A helper that shields an async iterable from its consumer's timing,
    and skips any values that the consumer is too slow to retrieve.

    Async iterables are only executed, like any Python generator, when it is queried (using `await __anext__()`).
    So, consumers that block between queries influence streams' timings.
    This context manager constantly queries the input stream in a task, i.e. without pauses between `__anext__` calls),
    in the process discarding any values the consumer was too slow to retrieve.
    Only one element is buffered for the consumer. Consider this order of operations:

        in: 0         # value 0 arrives on the input stream
        retrieve 0    # buffered value is retrieved by the consumer, returning immediately
        retrieve ...  # retrieving blocks
        in: 1
             ... 1    # retrieving returns
        in: 2
        in: 3         # previous value 2 is discarded
        retrieve 3
        in: 4
        in: EOF
        retrieve 4    # when EOF is reached, the last value remains in the buffer
        retrieve EOF

    Cancelling `await __anext__()` raises `Cancelled` into the iterator, thus terminating it.
    Therefore, another interface is used: the context manager resolves to a function that returns a tuple
    `(value, eof)`; either `(<value>, False)` or `(None, True)`.
    On exiting the context manager, the input stream is closed, thus cleaning up properly. Usage:

        async with __skipping_stream(stream) as anext:
            value, eof = await anext()
            while not eof:
                foo(value)
                value, eof = await anext()
    """

    async with trio.open_nursery() as nursery:
        # has the input stream emitted a value, or reached EOF, since last looking?
        news = trio.Event()
        # has there been a value from the input stream since last emitting one to the consumer?
        has_value = False
        # has the input stream reached EOF? This may be the case even if `has_value` is also true.
        eof = False
        # what's the last value emitted by the input stream? Only valid if `has_value` is true.
        value = None

        @nursery.start_soon
        async def reader():
            nonlocal has_value, eof, value
            try:
                async for value in stream:
                    # there's a value available in `value`
                    has_value = True
                    news.set()
                eof = True
                news.set()
            finally:
                if hasattr(stream, 'aclose'):
                    await stream.aclose()
                else:  # pragma: nocover
                    pass

        async def anext():
            nonlocal has_value, eof, value

            # If `eof`, there will be no events. `has_value` may be true or not, but any input is done.
            # If `not eof`, there *will* be more events.
            # - if `has_value` is set, the event will also be set. In that case waiting is a no-op and thus fine
            # - if `has_value` is not set, then we have to wait - either for EOF or for a value
            if not eof:
                await news.wait()
                news.clear()

            # now there's news, may be a value, may be EOF (or both)
            if has_value:
                has_value = False
                return value, False
            elif eof:
                return None, True
            else:  # pragma: nocover
                assert False

        yield anext
        nursery.cancel_scope.cancel()


async def subscription_transform(stream: AsyncIterator[T], timeout: float=None,
        granularity: Callable[[T, T], bool]=None, granularity_timeout: float=None) -> AsyncIterator[T]:
    """\
    Implements the stream transformation described in `subscription.proto`.
    The identity transform would be `subscription_transform(stream, granularity=lambda a, b: True)`:
    no timing behavior is added, and all values are treated as distinct, and thus emitted.

    If `granularity` is not given, values are compared for equality, thus from `[0, 0, 2, 1, 1, 0, 0, 1]`,
    elements 1, 4, and 6 would be discarded as being duplicates of their previous values.
    A typical example granularity measure for numbers is a lower bound on value difference,
    e.g. `lambda a, b: abs(a-b) > THRESHOLD`.

    The `timeout` parameter specifies a minimum time to pass between subsequent emitted values.
    After the timeout has passed, the most recently received value (if any) will be considered
    as if it had just arrived on the input stream,
    and then all subsequent values are considered until the next emission.
    Suppose the input is [0, 1, 0, 1, 0] and the timeout is just enough to skip one value completely.
    After emitting `0`, the first `1` is skipped, and the second `0` is not emitted because it's not a new value.
    The second `1` is emitted; because at that time no timeout is active (the last emission was too long ago).
    Immediately after the emission the timeout starts again, the last `0` arrives and the input stream ends.
    Because the `0` should be emitted, the stream awaits the timeout a final time, emits the value, and then terminates.
    Had the last value been a `1`, the output stream would have terminated immediately,
    as the value would not be emitted.

    The `granularity_timeout` parameter specifies a maximum time to pass between subsequent emitted values,
    as long as there were input values at all.
    The `granularity` may discard values of the input stream,
    leading in the most extreme case to no emitted values at all.
    If a `granularity_timeout` is given, then the most recent input value is emitted after that time,
    restarting both the ordinary and granularity timeout in the process.
    Suppose the input is [0, 0, 0, 1, 1, 0, 0] and the granularity timeout is just enough to skip one value completely.
    After emitting `0` and skipping the next one, another `0` is emitted:
    although the default granularity discarded the unchanged value, the granularity timeout forces its emission.
    Then, the first `1` and next `0` are emitted as normal, as changed values appeared before the timeout ran out.
    After the last `0`, the input ends.
    The stream waits a final time for the granularity timeout, outputs the value, and then terminates.

    Suppose the input is [0, 0] and the granularity timeout is so low that it runs out before the second zero.
    The first zero is the last value seen before the granularity timeout ran out,
    but once emitted it is out of the picture. The second zero is simply emitted as soon as it arrives.
    """
    if timeout is None:
        timeout = -math.inf
    if granularity is None:
        granularity = lambda a, b: a != b
    if granularity_timeout is None:
        granularity_timeout = math.inf

    async with _skipping_stream(stream) as anext:
        # we need a first value for our granularity checks
        value, eof = await anext()
        if eof:
            return

        while True:
            last_emit_at = trio.current_time()
            yield value
            if eof:
                return

            last_value = value
            has_value = False

            with trio.move_on_at(last_emit_at + granularity_timeout):
                with trio.move_on_at(last_emit_at + timeout):
                    while not eof:
                        # wait until there's news, save a value if there is
                        _value, eof = await anext()
                        if not eof:
                            # there's a value
                            value = _value
                            has_value = True

                # if we get here, either the timeout ran out or EOF was reached
                if eof:
                    if not has_value:
                        # no value at all
                        return
                    elif granularity(last_value, value):
                        # a good value! Wait for the timeout, then emit that value
                        await trio.sleep_until(last_emit_at + timeout)
                        continue
                    elif granularity_timeout < math.inf:
                        # there's still a chance to send the value after the granularity timeout
                        await trio.sleep_forever()
                    else:
                        # again, nothing to send
                        return
                    # note that none of the branches continue here

                # not EOF, so do regular waiting for values
                while not eof and (not has_value or not granularity(last_value, value)):
                    # wait until there's news, save a value if there is
                    _value, eof = await anext()
                    if not eof:
                        # there's a value
                        value = _value
                        has_value = True

                if eof:
                    # EOF was reached.
                    # If there is a value, we know that the granularity did not break the loop;
                    # no need to check that again.
                    if not has_value:
                        # no value at all
                        return
                    elif granularity_timeout < math.inf:
                        # there's still a chance to send the value after the granularity timeout
                        await trio.sleep_forever()
                    else:
                        # again, nothing to send
                        return
                    # note that none of the branches continue here

            # after the granularity timeout, we're fine with any value
            if has_value:
                continue

            # wait for the next event
            value, eof = await anext()
            if eof:
                return


class SubscriptionStreamer(Generic[T]):
    """
    `SubscriptionStreamer` implements the behavior regarding timeout, granularity, and granularity timeout
    described in subscription.proto.

    SubscriptionStreamer receives updates via `send` and `close`
    and forwards them to all output streams created with `subscribe`, if there are any.
    Each output stream then assesses whether and when to yield the update value, according to its parameters.

    A closed output stream will no longer receive items, and when `close` is called,
    all output streams will eventually terminate as well.
    """

    _EOF = object()

    def __init__(self) -> None:
        self.broadcast = BroadcastChannel()

    async def send(self, item: T) -> None:
        await self.broadcast.send(item)

    async def close(self) -> None:
        await self.broadcast.aclose()

    def subscribe(self, timeout: float=None,
                  granularity: Callable[[T, T], bool]=None, granularity_timeout: float=None) -> AsyncIterator[T]:
        return subscription_transform(self.broadcast.add_receiver(10), timeout, granularity, granularity_timeout)


# TODO how to extract logic from this so that it is not coupled to the Hedgehog server?

class Subscribable(Generic[T, Upd]):
    def __init__(self) -> None:
        self.streamer = SubscriptionStreamer[T]()
        self.subscriptions = {}  # type: Dict[Header, SubscriptionHandle]

    def compose_update(self, server: HedgehogServer, ident: Header, subscription: Subscription, value: T) -> Upd:
        raise NotImplemented

    async def subscribe(self, server: HedgehogServer, ident: Header, subscription: Subscription) -> None:
        raise NotImplemented


class SubscriptionHandle(object):
    def __init__(self, server: HedgehogServer, update_sender) -> None:
        self.count = 0
        self.server = server
        self._update_sender = update_sender
        self._scope = None

    async def _update_task(self, *, task_status=trio.TASK_STATUS_IGNORED):
        with trio.CancelScope() as scope:
            task_status.started(scope)
            await self._update_sender()

    async def increment(self) -> None:
        if self._scope is not None:
            self._scope.cancel()
        self._scope = await self.server.add_task(self._update_task)
        self.count += 1

    async def decrement(self) -> None:
        self.count -= 1
        if self.count == 0:
            self._scope.cancel()
            self._scope = None


class TriggeredSubscribable(Subscribable[T, Upd]):
    """
    Represents a value that changes by triggers known to the server, so it doesn't need to be actively polled.
    """

    def __init__(self) -> None:
        super(TriggeredSubscribable, self).__init__()

    async def update(self, value: T) -> None:
        await self.streamer.send(value)

    async def _update_sender(self, server: HedgehogServer, ident: Header, subscription: Subscription):
        async for value in self.streamer.subscribe(subscription.timeout / 1000):
            async with server.job():
                await server.send_async(ident, self.compose_update(server, ident, subscription, value))

    async def subscribe(self, server: HedgehogServer, ident: Header, subscription: Subscription) -> None:
        # TODO incomplete
        key = (ident, subscription.timeout)

        if subscription.subscribe:
            if key not in self.subscriptions:
                update_sender = partial(self._update_sender, server, ident, subscription)
                handle = self.subscriptions[key] = SubscriptionHandle(server, update_sender)
            else:
                handle = self.subscriptions[key]

            await handle.increment()
        else:
            try:
                handle = self.subscriptions[key]
            except KeyError:
                raise FailedCommandError("can't cancel nonexistent subscription")
            else:
                await handle.decrement()
                if handle.count == 0:
                    del self.subscriptions[key]


class PolledSubscribable(Subscribable[T, Upd]):
    """
    Represents a value that changes by independently from the server, so it is polled to observe changes.
    """

    def __init__(self) -> None:
        super(PolledSubscribable, self).__init__()
        self.intervals: trio.abc.SendChannel = None
        self.timeouts = set()  # type: Set[float]

    @property
    def _registered(self):
        return self.intervals is not None

    async def poll(self) -> T:
        raise NotImplemented

    async def _poll_task(self, *, task_status=trio.TASK_STATUS_IGNORED):
        self.intervals, intervals = trio.open_memory_channel(0)
        task_status.started()

        async with trio.open_nursery() as nursery:
            async def poller(interval, *, task_status=trio.TASK_STATUS_IGNORED):
                with trio.CancelScope() as scope:
                    task_status.started(scope)
                    while True:
                        await self.streamer.send(await self.poll())
                        await trio.sleep(interval)

            @nursery.start_soon
            async def read_interval():
                scope = None
                async for interval in intervals:
                    # cancel the old polling task
                    if scope is not None:
                        scope.cancel()
                        scope = None

                    # start new polling task
                    if interval >= 0:
                        scope = await nursery.start(poller, interval)

    async def _update_sender(self, server: HedgehogServer, ident: Header, subscription: Subscription):
        async for value in self.streamer.subscribe(subscription.timeout / 1000):
            async with server.job():
                await server.send_async(ident, self.compose_update(server, ident, subscription, value))

    async def register(self, server: HedgehogServer) -> None:
        if not self._registered:
            await server.add_task(self._poll_task)

    async def subscribe(self, server: HedgehogServer, ident: Header, subscription: Subscription) -> None:
        await self.register(server)

        # TODO incomplete
        key = (ident, subscription.timeout)

        if subscription.subscribe:
            if key not in self.subscriptions:
                self.timeouts.add(subscription.timeout / 1000)
                update_sender = partial(self._update_sender, server, ident, subscription)
                handle = self.subscriptions[key] = SubscriptionHandle(server, update_sender)
            else:
                handle = self.subscriptions[key]

            await handle.increment()
            await self.intervals.send(min(self.timeouts))
        else:
            try:
                handle = self.subscriptions[key]
            except KeyError:
                raise FailedCommandError("can't cancel nonexistent subscription")
            else:
                await handle.decrement()
                if handle.count == 0:
                    self.timeouts.remove(subscription.timeout / 1000)
                    await self.intervals.send(min(self.timeouts, default=-1))

                    del self.subscriptions[key]
